from typing import TYPE_CHECKING
import pytest
import asyncio
import random

from simulation.location import Location
from simulation.action import PickupAction
from simulation.event import PickedUpEvent
from simulation.interactables.pickups import Artefact


from .mock_world import MockWorld
from tests.test_simulation.dummy_avatar import CustomLiveDummy

if TYPE_CHECKING:
    from simulation.cell import Cell
    from simulation.avatar.avatar_wrapper import AvatarWrapper


@pytest.fixture
def game() -> "MockWorld":
    world = MockWorld(dummies_list=[CustomLiveDummy])
    world.simulation_runner.add_avatar(1, Location(0, 0))
    return world


@pytest.fixture
def cell(game):
    return game.game_state.world_map.get_cell(Location(1, 0))


def test_artefact_serialization(cell):
    artefact = Artefact(cell)
    assert artefact.serialize() == {
        "type": "artefact",
        "location": {"x": cell.location.x, "y": cell.location.y},
    }


@pytest.mark.asyncio
async def test_artefact_applies_correctly(game, cell):
    avatar: "CustomLiveDummy" = game.avatar_manager.get_avatar(1)
    cell.interactable = Artefact(cell)

    # Move to the cell with the artefact

    await game.simulation_runner.run_single_turn(
        game.avatar_manager.get_player_id_to_serialized_action()
    )

    assert cell.interactable is not None

    # Pickup the artefact

    avatar.set_next_action(PickupAction(avatar))

    await game.simulation_runner.run_single_turn(
        game.avatar_manager.get_player_id_to_serialized_action()
    )

    assert cell.avatar == avatar
    assert cell.interactable is None
    assert avatar.number_of_artefacts == 1



from simulation.worker_manager import WorkerManager
from simulation.workers.worker import Worker


class MockWorker(Worker):
    async def fetch_data(self, state_view):
        pass

    def _create_worker(self):
        pass


class TestWorkerManager:
    def set_up(self):
        self.worker_manager = WorkerManager()

    async def test_fetch_worker_data_is_async(self):
        """
        Creates a mock Worker and calls fetch_all_worker_data to check it returns the
        correct type. The type expected is a list, which is returned when the
        fetch_data function is correctly awaited in the worker manager, as opposed to
        the generator Class it returns if it not awaited properly.
        """
        self.set_up()
        self.worker_manager.player_id_to_worker[0] = MockWorker(1, 0000)
        data = await self.worker_manager.fetch_all_worker_data(
            self.worker_manager.player_id_to_worker
        )
        assert type(data) == list



from simulation.interactables.conditions import avatar_on_cell, pickup_action_applied
from simulation.interactables.effects import ArtefactEffect
from simulation.interactables.interactable import _Interactable


class Artefact(_Interactable):
    def __init__(self, cell):
        super(Artefact, self).__init__(cell)
        self.delete_after_effects_applied = True
        self.pickup_action_applied = False
        self.conditions = [avatar_on_cell, pickup_action_applied]
        self.effects.append(ArtefactEffect)

    def get_targets(self):
        return [self.cell.avatar]

    def __repr__(self):
        return "Artefact(Location={})".format(self.cell.location)

    def serialize(self):
        return {
            "type": "artefact",
            "location": {"x": self.cell.location.x, "y": self.cell.location.y},
        }



import asyncio
import logging
import os
from asyncio import Future
from concurrent import futures
from typing import Tuple

from simulation.workers import WORKER

LOGGER = logging.getLogger(__name__)
WORKER_TIMEOUT_TIME_SECONDS = 1


class WorkerManager(object):
    """
    Methods of this class must be thread safe unless explicitly stated.
    """

    def __init__(self, port=5000):
        self.player_id_to_worker = {}
        self.port = port
        self.worker_class = WORKER[os.environ.get("WORKER", "local")]

    def get_code(self, player_id):
        return self.player_id_to_worker[player_id].code

    async def fetch_all_worker_data(self, player_id_to_game_state):
        """
        Creates a thread for each worker to send a request for their data. After
        a set duration these threads will close, giving a consistent turn time.
        """
        worker_game_states = [
            (self.player_id_to_worker[player_id], player_id_to_game_state[player_id])
            for player_id in player_id_to_game_state.keys()
        ]

        requests = [
            worker.fetch_data(game_state) for worker, game_state in worker_game_states
        ]

        try:
            return await asyncio.wait_for(
                asyncio.gather(*requests, return_exceptions=True), WORKER_TIMEOUT_TIME_SECONDS
            )
        except futures.TimeoutError:
            LOGGER.warning("Fetching workers data timed out")

    def get_player_id_to_serialized_actions(self):
        return {
            player_id: self.player_id_to_worker[player_id].serialized_action
            for player_id in self.player_id_to_worker
        }

    def clear_logs(self):
        for worker in self.player_id_to_worker.values():
            worker.log = None

    def update_code(self, player):
        self.player_id_to_worker[player["id"]].code = player["code"]

    def add_new_worker(self, player_id):
        self.player_id_to_worker[player_id] = self.worker_class(player_id, self.port)

    async def _parallel_map(self, func, iterable_args):
        loop = asyncio.get_event_loop()
        with futures.ThreadPoolExecutor() as executor:
            workers: Tuple[Future] = (
                loop.run_in_executor(executor, func, args) for args in iterable_args
            )
            await asyncio.gather(*workers)

    async def add_workers(self, users_to_add):
        await self._parallel_map(self.add_new_worker, users_to_add)

    async def delete_workers(self, players_to_delete):
        await self._parallel_map(self.delete_worker, players_to_delete)

    def delete_worker(self, player_id):
        if player_id in self.player_id_to_worker:
            worker = self.player_id_to_worker[player_id]
            del self.player_id_to_worker[player_id]
            worker.remove_worker()

    async def update_worker_codes(self, players):
        await self._parallel_map(self.update_code, players)
